# encoding: utf8 #
'''
This module contains the base class for the metrics that provide the
base functions to register/unregister counter and also pack/unpack data.
'''
from functools import reduce
from datetime import datetime
from colmet.common.exceptions import CounterAlreadyExistError, NoneValueError


import struct
import operator
import ctypes

######################
# Base counter types #
######################


class BaseType(object):
    struct_code = None
    length = None

    before_pack = lambda self, value: value
    after_unpack = lambda self, value: value

    def __eq__(self, other):
        return (self.struct_code == other.struct_code
                and self.length == other.length)


class UInt8(BaseType):
    struct_code = 'H'
    length = struct.calcsize("<%s" % struct_code)


class UInt16(BaseType):
    struct_code = 'H'
    length = struct.calcsize("<%s" % struct_code)


class UInt32(BaseType):
    struct_code = 'I'
    length = struct.calcsize("<%s" % struct_code)


class UInt64(BaseType):
    struct_code = 'Q'
    length = struct.calcsize("<%s" % struct_code)


class UFloat(BaseType):
    struct_code = 'f'
    length = struct.calcsize("<%s" % struct_code)


class UDouble(BaseType):
    struct_code = 'd'
    length = struct.calcsize("<%s" % struct_code)


class String(BaseType):
    def __init__(self, length):
        self.struct_code = '%ss' % length
        self.length = struct.calcsize("<%s" % self.struct_code)

    after_unpack = lambda self, value: value.rstrip("\0")


#####################
# Base metric class #
#####################


class MetaCountersType(type):
    '''
    This type handle the counters/headers registration for the metric. For each
    counter/header defined in the list _counters/_headers, it create a
    corresponding property and register their specifities for the class usage.
    '''

    '''
    Define the representation of the counters.
    '''
    _counter_representations = {
        'bytes': lambda value: "%s (bytes)" % MetaCountersType._normalize(value, 1024, 10000, ['', 'K', 'M', 'G', 'T']),
        'kbytes': lambda value: "%s (bytes)" % MetaCountersType._normalize(value, 1024, 10000, ['K', 'M', 'G', 'T']),
        "mbytes": lambda value: "%s (bytes)" % MetaCountersType._normalize(value, 1024, 10000, ['M', 'G', 'T']),
        "ts_date": lambda value: datetime.fromtimestamp(value).strftime("%d/%m/%Y - %H:%M:%S"),
        "sec": lambda value: "%s (seconds)" % str(value),
        "usec": lambda value: "%s (seconds)" % MetaCountersType._normalize(value, 1000, 1000, ['u', 'm', '']),
        "nsec": lambda value: "%s (seconds)" % MetaCountersType._normalize(value, 1000, 1000, ['n', 'u', 'm', '']),
        "count": lambda value: "%s (count)" % MetaCountersType._normalize(value, 1000, 10000, ['', 'K', 'M', 'G', 'T']),
        "mbytes-usec": lambda value: "%s (bytes*seconds)" % MetaCountersType._normalize(float(value * 1024 * 1024 / 1000 / 1000), 1024, 10000, [' ', 'K', 'M', 'G', 'T']),
        "n/a": lambda value: "%s" % str(value)
    }

    @staticmethod
    def _normalize(value, factor, limit, exts):
        '''
        Return the normalized 'value' according to 'factor', 'limit' and
        the extentions given in 'exts'. Until 'value' is bigger than 'limit',
        'value' is divided by 'factor'.
        '''
        exp = 0

        try:
            val = float(value)
        except TypeError:
            # value is not available (in case of newly created cgroup with no tasks)
            raise NoneValueError

        while val > limit and exp < len(exts) - 1:
            val = val / factor
            exp += 1
        return "%.0f%s" % (val, exts[exp])

    '''
    The accumulation functions used for the counters. These functions
    takes 3 parameters: x, y and a coefficient.

    By default there are the functions:
    - 'add'  : x,y,coeff: y if x == None else x if y == None else x + coeff * y
    - 'none' : x,y,coeff -> y if x == None else y
    - 'min'  : x,y,coeff -> y if x == None else x if y == None else x if x < y else y
    - 'max'  : x,y,coeff -> y if x == None else x if y == None else x if x >= y else y
    '''
    _counter_accumulation_functions = {
        'add': lambda x, y, coeff: y if x is None else x if y is None else x + coeff * y,
        'none': lambda x, y, coeff: y if x is None else x,
        'min': lambda x, y, coeff: y if x is None else x if y is None else x if x < y else y,
        'max': lambda x, y, coeff: y if x is None else x if y is None else x if x >= y else y,
    }

    def __new__(cls, name, bases, attrs):
        if len(bases) > 1:
            raise TypeError("Counters cannot be derived from multiple class")

        attrs['_header_definitions'] = {}
        attrs['_counter_definitions'] = {}
        attrs['_fmt'] = None
        attrs['_counter_representations'] = MetaCountersType._counter_representations
        attrs['_counter_accumulation_functions'] = MetaCountersType._counter_accumulation_functions

        # Copy the header/counter definition from the parent class if
        # any
        if len(bases) == 1:
            parent_cls = bases[0]
            for key in ['_header_definitions', '_counter_definitions', '_fmt',
                        '_counter_represensations',
                        '_counter_accumulation_functions']:
                if hasattr(parent_cls, key):
                    parent_item = getattr(parent_cls, key)
                    if isinstance(parent_item, (dict, set)):
                        attrs[key] = parent_item.copy()
                    else:
                        attrs[key] = parent_item

        return type.__new__(cls, name, bases, attrs)

    def __init__(self, name, bases, attrs):
        # Registering headers
        if '_headers' in attrs:
            for (h_name, h_type, h_repr) in attrs['_headers']:
                self.register_header(h_name, h_type, h_repr)
                self._set_header_property(h_name)

        # Registering counters
        if '_counters' in attrs:
            for c_name, c_type, c_repr, c_acc_fn, c_descr in attrs['_counters']:
                self.register_counter(c_name, c_type, c_repr, c_acc_fn, c_descr)
                self._set_counter_property(c_name)

        return type.__init__(self, name, bases, attrs)

    def _set_header_property(self, name):
        '''
        Internal function to set the property setter/getter for a given header
        value
        '''
        fget = lambda t: t._get_header(name)
        fset = lambda t, value: t._set_header(name, value)
        setattr(self, name, property(fget, fset))

    def _set_counter_property(self, name):
        '''
        Internal function to set the property setter/getter for a given counter
        value
        '''
        fget = lambda t: t._get_counter(name)
        fset = lambda t, value: t._set_counter(name, value)
        setattr(self, name, property(fget, fset))

    def register_counter(self, c_name, c_type, c_representation,
                         c_accumulate_function, c_descr, c_index=None):
        '''
        Register a counter for the given class.

        This method must be used as a class method when defining a derived
        class of BaseCounters
        '''
        if c_name in self._counter_definitions or c_name in self._header_definitions:
            raise CounterAlreadyExistError()

        if c_index is None:
            if len(self._counter_definitions) > 0:
                c_index = 1 + max([index for (_, (_, _, _, _, index, _)) in
                                  self._counter_definitions.iteritems()])
            else:
                c_index = 0

        self._counter_definitions[c_name] = (
            c_type,
            c_representation,
            c_accumulate_function,
            c_descr,
            c_index,
            None
        )

        self._update_struct_fmt()

    def register_header(self, h_name, h_type, h_representation, h_index=None):
        '''
        Register a header for the given class

        This method must be used as a class method when defining a derived
        class of BaseCounters
        '''
        if h_name in self._header_definitions or h_name in self._counter_definitions:
            raise CounterAlreadyExistError()

        if h_index is None:
            if len(self._header_definitions) > 0:
                h_index = 1 + max([index for (_, (_, _, index, _))
                                  in self._header_definitions.iteritems()])
            else:
                h_index = 0

        self._header_definitions[h_name] = (
            h_type,
            h_representation,
            h_index,
            None
        )
        self._update_struct_fmt()

    def _update_struct_fmt(self):
        '''
        Internal method to update the struct format of the class.

        the struct format is computed from the _headers and _counters definitions.
        '''
        offset = 0
        sorted_hdef = sorted(self._header_definitions.items(), key=lambda v: v[1][2])
        for (h_name, (h_type, h_representation, h_index, _)) in sorted_hdef:
            self._header_definitions[h_name] = (h_type, h_representation, h_index, offset)
            offset += h_type.length
        sorted_cdef = sorted(self._counter_definitions.items(), key=lambda v: v[1][4])
        for (c_name, (c_type, c_representation, c_acc_fn, c_descr, c_index, _)) in sorted_cdef:
            self._counter_definitions[c_name] = (c_type, c_representation, c_acc_fn, c_descr, c_index, offset)
            offset += c_type.length

        sorted_hdef = sorted(self._header_definitions.items(), key=lambda v: v[1][2])
        header_fmt_list = [
            (h_name, h_type.struct_code)
            for (h_name, (h_type, _, _, _)) in sorted_hdef
        ]
        sorted_cdef = sorted(self._counter_definitions.items(), key=lambda v: v[1][4])
        counter_fmt_list = [
            (c_name, c_type.struct_code)
            for (c_name, (c_type, _, _, _, _, _)) in sorted_cdef
        ]

        h_key_list, h_fmt_code_list = zip(*header_fmt_list) if len(header_fmt_list) > 0 else ([], [])
        c_key_list, c_fmt_code_list = zip(*counter_fmt_list) if len(counter_fmt_list) > 0 else ([], [])

        self._fmt = "<" + "".join(list(h_fmt_code_list)) + "".join(list(c_fmt_code_list))
        self._fmt_length = struct.calcsize(self._fmt)

        self._fmt_header_ordered_keys = list(h_key_list)
        self._fmt_counter_ordered_keys = list(c_key_list)


class BaseCounters(object):
    '''
    This class define the base class for a metric. It provides the basic
    functions for packing/unpacking, registering header/counter.

    The counters and headers are registered by adding an entry into
    the lists '_headers' and '_counters'.

    By default the headers contain the 'metric_backend', the 'hostname',
    the 'job_id', and the 'timestamp'.  '''

    __metaclass__ = MetaCountersType
    __metric_name__ = "base"
    _headers = [('metric_backend', String(255), 'string'),
                ('hostname', String(255), 'string'),
                ('job_id', UInt64(), 'count'),
                ('timestamp', UInt64(), 'count')]

    _counters = [
    ]

    ################
    # Static methods
    ################

    @staticmethod
    def create_metric_from_raw(raw):
        from . import get_counters_class
        backend = struct.unpack('255s', raw[0:255])[0].rstrip("\0")
        counters_class = get_counters_class(backend)
        counters = counters_class(raw=raw[0:counters_class._fmt_length])
        return counters

    @staticmethod
    def pack_from_list(counters_list):
        length = reduce(operator.add, [counters._fmt_length for counters in counters_list])
        raw = ctypes.create_string_buffer(length)

        offset = 0
        for counters in counters_list:
            new_offset = offset + counters._fmt_length
            counters.pack_into(raw, offset)
            offset = new_offset
        return raw

    @staticmethod
    def unpack_to_list(raw, unpack_counters=False):
        length = len(raw)
        offset = 0
        counters_list = list()
        while(offset < length):
            counters = BaseCounters.create_metric_from_raw(raw[offset:])
            offset += counters._fmt_length
            counters_list.append(counters)

        if unpack_counters:
            for counters in counters_list:
                counters.unpack()

        return counters_list

    #################
    # Class methods #
    #################
    @classmethod
    def get_zero_counters(cls, timestamp=None, *args):
        raise TypeError(
            "The metrics need to implement "
            "how to get zero counters"
        )

    @classmethod
    def fetch(cls, input_backend, timestamp, *args):
        raise TypeError(
            "The metrics need to implemend "
            "how to get counters"
        )

    ##################
    # Object methods #
    ##################
    #
    # Private object methods
    #
    def _is_buffer(self):
        return self._buf is not None

    def _create_buffer(self):
        self._buf = ctypes.create_string_buffer(self._fmt_length)
        return self._buf

    def _empty_fill(self):
        if self._packed:
            self._buf.raw.zfill(self._fmt_length)
        else:
            for c_name in self._counter_definitions.keys():
                self._set_counter(c_name, None)

    def _set_header(self, key, value):
        if self._packed:
            (h_type, _, _, h_offset) = self._header_definitions[key]
            struct.pack_into(h_type.struct_code, self._buf, h_offset, h_type.before_pack(value))
        else:
            self._header_values[key] = value

    def _get_header(self, key):
        if self._packed:
            (h_type, _, _, h_offset) = self._header_definitions[key]
            return h_type.after_unpack(struct.unpack_from(h_type.struct_code, self._buf, h_offset)[0])
        else:
            return self._header_values[key]

    def _get_counter(self, key):
        if self._packed:
            (c_type, _, _, _, _, c_offset) = self._counter_definitions[key]
            return c_type.after_unpack(struct.unpack_from(c_type.struct_code, self._buf, c_offset)[0])
        else:
            return self._counter_values[key]

    def _set_counter(self, key, value):
        if self._packed:
            (c_type, _, _, _, _, c_offset) = self._counter_definitions[key]
            struct.pack_into(c_type.struct_code, self._buf, c_offset, c_type.before_pack(value))
        else:
            self._counter_values[key] = value

    def __repr__(self):
        msg = "%s  Job(%s)/ %s -> Host(%s) -> %s" % (
            self.timestamp,
            self.job_id,
            self.metric_backend,
            self.hostname,
            self._format_counters()
        )
        return msg

    def _format_counters(self, prefix=""):
        '''
        Return a string formatted that represent the counters 'prefix' is inserted to the formatted string.
        '''
        msg_counters = {}
        key_maxlen = 0
        for key in self._counter_definitions.keys():
            if len(key) > key_maxlen:
                key_maxlen = len(key)
        msg_format_values = "\n\t" + prefix + "%" + str(key_maxlen) + "s : %s"
        for k in self._counter_definitions.keys():
            (_, c_repr, _, c_index, _, _) = self._counter_definitions[k]
            if c_index not in msg_counters:
                msg_counters[c_index] = ""

            msg_counters[c_index] += msg_format_values % (
                k,
                self._counter_representations[c_repr](self._get_counter(k))
            )
        sorted_msg = [msg_counters[k] for k in sorted(msg_counters.keys())]
        return (prefix + '{%s\n' + prefix + '}') % ("".join(sorted_msg))

    #
    # Public object methods
    #

    def get_metric_name(self):
        return self.__metric_name__

    def pack(self):
        '''
        Convert the data into the packed form
        '''
        if not self._is_buffer():
            self._create_buffer()

        self.pack_into(self._buf)
        self._packed = True

    def unpack(self):
        '''
        Convert the data into the unpacked form
        '''
        if not self._is_buffer():
            raise ValueError("The metric buffer should not be empty")

        self.unpack_from(self._buf)
        self._packed = False

    def pack_into(self, raw_buffer, offset=0):
        '''
        Convert the data into the packed form to a specific buffer
        '''

        fmt_values = ([self._header_definitions[key][0].before_pack(self._get_header(key)) for key in self._fmt_header_ordered_keys]
                      + [self._counter_definitions[key][0].before_pack(self._get_counter(key)) for key in self._fmt_counter_ordered_keys])
        struct.pack_into(self._fmt, raw_buffer, offset, *fmt_values)

    def unpack_from(self, raw_buffer, offset=0):
        '''
        Convert the data into the packed form from a specific buffer
        '''
        fmt_values = struct.unpack_from(self._fmt, raw_buffer, offset)
        index = 0
        for key in self._fmt_header_ordered_keys:
            self._header_values[key] = self._header_definitions[key][0].after_unpack(fmt_values[index])
            index += 1
        for key in self._fmt_counter_ordered_keys:
            self._counter_values[key] = self._counter_definitions[key][0].after_unpack(fmt_values[index])
            index += 1

    def get_packed(self):
        if not self._packed:
            self.pack()
        return self._buf[0:self._fmt_length]

    def __init__(self, raw=None):
        self._counter_values = {}
        self._header_values = {}
        if raw is not None:
            if len(raw) != self._fmt_length:
                raise ValueError(
                    "The buffer length (%s) given in parameters "
                    "must be the same size as the metric (%s)"
                    % (len(raw), self._fmt_length)
                )
            self._packed = True
            self._buf = ctypes.create_string_buffer(raw)
        else:
            self._packed = False
            self._buf = None
            self.metric_backend = self.get_metric_name()

    def accumulate(self, other_stats, destination, coeff=1):
        """Update destination from operator(self, other_stats)"""
        for counters in [self, other_stats, destination]:
            if counters._packed:
                counters.unpack()

        d_counters = destination._counter_values
        o_counters = other_stats._counter_values
        s_counters = self._counter_values
        for (name, (_, _, acc_fn, _, _, _)) in self._counter_definitions.iteritems():
            d_counters[name] = \
                self._counter_accumulation_functions[acc_fn](s_counters[name],
                                                             o_counters[name],
                                                             coeff)

    def delta(self, other_stats, destination):
        """Update destination with self - other_stats"""
        return self.accumulate(other_stats, destination, coeff=-1)



# -*- coding: utf-8 -*-
"""
Testing the base metric class
"""
import pytest

from colmet.common.metrics.base import (UInt8, UInt16, UInt32, UInt64, String,
                                        BaseCounters, CounterAlreadyExistError)


def test_basic_types():
    assert UInt8().length == 2
    assert UInt16().length == 2
    assert UInt32().length == 4
    assert UInt64().length == 8
    for i in [1, 5, 0, 100]:
        assert String(i).length == i


class MyCounters(BaseCounters):
    _headers = [
        ('project', String(16), 'string')
    ]

    _counters = [
        ('speed', UInt64(), 'count', 'add', 'Speed')
    ]


@pytest.fixture(scope="function")
def mycounters(request, monkeypatch):
    m = MyCounters()
    m.job_id = 42
    m.timestamp = 0
    m.hostname = 'localhost'
    return m


def test_register_header():
    '''Testing registering header'''
    assert 'project' in MyCounters._header_definitions

    (h_type, h_repr, h_index, h_offset) = MyCounters._header_definitions[
        'project']

    assert h_type == String(16)
    assert h_repr == 'string'
    assert h_offset is not None
    assert h_index is not None


def test_register_counters():
    '''Testing registering counter'''
    assert 'speed' in MyCounters._counter_definitions

    class MyCounters2(MyCounters):

        _counters = [
            ('speed2', UInt32(), 'count', 'add', 'Speed2')
        ]

    assert 'speed2' in MyCounters2._counter_definitions

    (c_type, c_repr, c_acc_fn, c_descr, c_index,
     c_offset) = MyCounters2._counter_definitions['speed2']

    assert c_type == UInt32()
    assert c_repr == 'count'
    assert c_acc_fn == 'add'
    assert c_descr == 'Speed2'
    assert c_offset is not None
    assert c_index is not None


def test_counter_assignment(mycounters):
    '''Testing basic counters get/set'''
    mycounters.speed = 120
    assert(mycounters.speed == 120)
    assert(mycounters._get_counter('speed') == 120)


def test_counter_packed_assignment(mycounters):
    '''Testing packed/unpacked counters get/set'''
    mycounters.project = 'speedcontrol'
    mycounters.speed = 0
    mycounters.pack()
    assert(mycounters.speed == 0)
    mycounters.speed = 120
    assert(mycounters.speed == 120)
    mycounters.unpack()
    assert(mycounters.speed == 120)


def test_failure_if_regitering_an_existing_counter():
    '''Registering an existing counter should fail'''
    with pytest.raises(CounterAlreadyExistError):
        class MyCounters2(MyCounters):
            _counters = [
                ('speed', UInt64(), 'count', 'add', 'Speed')
            ]


def test_failure_if_regitering_an_existing_headerr():
    '''Registering an existing counter should fail'''
    with pytest.raises(CounterAlreadyExistError):
        class MyCounters2(MyCounters):
            _headers = [
                ('job_id', UInt64(), 'string')
            ]


def test_normalize():
    '''Testing formating value (Normalization)'''
    val = BaseCounters._normalize(1, 1000, 10000, ['', 'K', 'G'])
    assert val == "1"
    val = BaseCounters._normalize(1000, 1000, 10000, ['', 'K', 'G'])
    assert val == "1000"
    val = BaseCounters._normalize(100000, 1000, 10000, ['', 'K', 'G'])
    assert val == "100K"


def test_failure_if_deriving_counters_from_multiple_class():
    class MyCounters2(MyCounters):
        pass

    with pytest.raises(TypeError):
        class MyCounters3(MyCounters, MyCounters2):
            pass



import logging

LOG = logging.getLogger()

from .base import UInt64, BaseCounters


class LustrestatsCounters(BaseCounters):
    __metric_name__ = 'lustrestats_default'

    counters_lustrestats = {
        # 'key': ( offset,length, type, repr, acc )
        'lustre_nb_read': (UInt64(), 'count', 'none', 'lustre_nb_read'),
        'lustre_bytes_read': (UInt64(), 'bytes', 'none', 'lustre_bytes_read'),
        'lustre_nb_write': (UInt64(), 'count', 'none', 'lustre_nb_write'),
        'lustre_bytes_write': (UInt64(), 'bytes', 'none', 'lustre_bytes_write'),
        }
        
    counters_lustrestats_to_get = [
        'lustre_nb_read',
        'lustre_bytes_read',
        'lustre_nb_write',
        'lustre_bytes_write'
    ]

    _counters = []
        
    for c_name in counters_lustrestats_to_get:
        (c_type, c_repr, c_acc, c_descr) = counters_lustrestats[c_name]
        _counters.append((c_name, c_type, c_repr, c_acc, c_descr))

    @classmethod
    def get_zero_counters(cls):
        return cls(lustrestats_buffer=None, raw=None)

    @classmethod
    def fetch(cls, lustrestats_backend):
        return lustrestats_backend.get_lustrestats()

    def __init__(self, lustrestats_buffer=None, raw=None):
        BaseCounters.__init__(self, raw=raw)
        if raw is not None:
            pass
        elif lustrestats_buffer is None:
            self._empty_fill()
        else:
            for name in LustrestatsCounters._counter_definitions:
                self._counter_values[name] = lustrestats_buffer[name]



'''
ZeroMQ input backend for colmet collector
'''
import zmq
import logging
import socket
import struct


LOG = logging.getLogger()

from colmet.common.metrics.base import BaseCounters

from colmet.common.backends.base import InputBaseBackend, OutputBaseBackend

try:
    _snd_hwm = zmq.HWM
    _rcv_hwm = zmq.HWM
except AttributeError:
    _snd_hwm = zmq.SNDHWM
    _rcv_hwm = zmq.RCVHWM


class ZMQInputBackend(InputBaseBackend):
    __backend_name__ = "zeromq"

    def open(self):
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PULL)
        self.socket.setsockopt(zmq.LINGER, self.options.zeromq_linger)
        self.socket.setsockopt(_rcv_hwm, self.options.zeromq_hwm)
        LOG.debug("Use the bind URI '%s'" % self.options.zeromq_bind_uri)
        self.socket.bind(self.options.zeromq_bind_uri)

    def close(self):
        self.socket.close()
        self.context.term()

    def pull(self, buffer_size=1000):
        counters_list = []
        try:
            for i in xrange(buffer_size):
                raw = self.socket.recv(zmq.NOBLOCK, copy=False)
                counters_list.extend(BaseCounters.unpack_to_list(raw.bytes))
                del raw
        except zmq.ZMQError, e:
            if e.errno != zmq.EAGAIN:
                raise e
        LOG.debug("%s counters received" % len(counters_list))
        if len(self.job_id_list) > 0:
            counters_list = [metric for metric in counters_list
                             if metric.job_id in self.job_id_list]
            LOG.debug("%s counters received after filtering"
                      % len(counters_list))

        return counters_list


class ZMQOutputBackend(OutputBaseBackend):
    '''
    zmq node backend class
    '''
    __backend_name__ = "zeromq"

    def open(self):
        self.context = zmq.Context()
        self.hostname = socket.gethostname()
        self.socket = self.context.socket(zmq.PUSH)
        self.socket.setsockopt(zmq.LINGER, self.options.zeromq_linger)
        self.socket.setsockopt(_snd_hwm, self.options.zeromq_hwm)
        self.socket.connect(self.options.zeromq_uri)
        LOG.debug("Use the URI '%s'" % self.options.zeromq_uri)

    def close(self):
        self.socket.close()
        self.context.term()

    def push(self, counters_list):
        '''
        push the metrics to the output backend
        '''
        if len(counters_list) > 0:
            try:
                raw = BaseCounters.pack_from_list(counters_list)
                self.socket.send(raw)
            except struct.error as e:
                LOG.error("An error occurred during packet creation : %s" % e)



'''
Netlink message generation/parsing

Copyright 2007        Johannes Berg <johannes@sipsolutions.net>

GPLv2+; See copying for details.
'''

import struct
from netlink import NLM_F_REQUEST, NLMSG_MIN_TYPE, Message, parse_attributes
from netlink import NulStrAttr, Connection, NETLINK_GENERIC

CTRL_CMD_UNSPEC       = 0
CTRL_CMD_NEWFAMILY    = 1
CTRL_CMD_DELFAMILY    = 2
CTRL_CMD_GETFAMILY    = 3
CTRL_CMD_NEWOPS       = 4
CTRL_CMD_DELOPS       = 5
CTRL_CMD_GETOPS       = 6

CTRL_ATTR_UNSPEC      = 0
CTRL_ATTR_FAMILY_ID   = 1
CTRL_ATTR_FAMILY_NAME = 2
CTRL_ATTR_VERSION     = 3
CTRL_ATTR_HDRSIZE     = 4
CTRL_ATTR_MAXATTR     = 5
CTRL_ATTR_OPS         = 6

class GenlHdr:
    def __init__(self, cmd, version = 0):
        self.cmd = cmd
        self.version = version
    def _dump(self):
        return struct.pack("BBxx", self.cmd, self.version)

def _genl_hdr_parse(data):
    return GenlHdr(*struct.unpack("BBxx", data))

GENL_ID_CTRL        = NLMSG_MIN_TYPE

class GeNlMessage(Message):
    def __init__(self, family, cmd, attrs=[], flags=0):
        self.cmd = cmd
        self.attrs = attrs
        self.family = family
        Message.__init__(self, family, flags=flags,
                         payload=[GenlHdr(self.cmd)]+attrs)

    @staticmethod
    def recv(conn):
        msg = conn.recv()
        packet = msg.payload
        hdr = _genl_hdr_parse(packet[:4])

        genlmsg = GeNlMessage(msg.type, hdr.cmd, [], msg.flags)
        genlmsg.attrs = parse_attributes(packet[4:])
        genlmsg.version = hdr.version

        return genlmsg

class Controller:
    def __init__(self, conn):
        self.conn = conn
    def get_family_id(self, family):
        a = NulStrAttr(CTRL_ATTR_FAMILY_NAME, family)
        m = GeNlMessage(GENL_ID_CTRL, CTRL_CMD_GETFAMILY,
                        flags=NLM_F_REQUEST, attrs=[a])
        m.send(self.conn)
        m = GeNlMessage.recv(self.conn)
        return m.attrs[CTRL_ATTR_FAMILY_ID].u16()

connection = Connection(NETLINK_GENERIC)
controller = Controller(connection)



import os
import re

from colmet.common.backends.base import InputBaseBackend
from colmet.common.job import Job
from colmet.common.metrics.procstats import ProcstatsCounters


class ProcstatsBackend(InputBaseBackend):
    __backend_name__ = "procstats"

    def open(self):
        self.procstats = ProcStats(self.options)
        # job with id equal to 0 is the fictive job to gather nodes' monitoring
        # measures
        self.job_0 = Job(self, 0, self.options)

    def close(self):
        pass

    def get_procstats(self):
        counters = self.procstats.get_stats()
        return counters

    def get_counters_class(self):
        return ProcstatsCounters

    def pull(self):
        self.job_0.update_stats()
        return self.job_0.get_stats()

#
# The source part below is largely inspired by procstats.py file from open
# TSDB projet
#
# This file is part of tcollector.
# Copyright (C) 2010  StumbleUpon, Inc.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.  You should have received a copy
# of the GNU Lesser General Public License along with this program.  If not,
# see <http://www.gnu.org/licenses/>.

NUMADIR = "/sys/devices/system/node"


class ProcStats(object):

    def __init__(self, option):
        self.options = option

        self.f_uptime = open("/proc/uptime", "r")
        self.f_meminfo = open("/proc/meminfo", "r")
        self.f_vmstat = open("/proc/vmstat", "r")
        self.f_stat = open("/proc/stat", "r")
        self.f_loadavg = open("/proc/loadavg", "r")
        self.f_entropy_avail = open("/proc/sys/kernel/random/entropy_avail", "r")
#       self.numastats = self.open_sysfs_numa_stats()

# NOT USED TODO
    def open_sysfs_numa_stats():
        """Returns a possibly empty list of opened files."""
        try:
            nodes = os.listdir(NUMADIR)
        except OSError as e:
            if e.errno == 2:  # No such file or directory
                return []   # We don't have NUMA stats.
            raise

        nodes = [node for node in nodes if node.startswith("node")]
        numastats = []
        for node in nodes:
            try:
                numastats.append(open(os.path.join(NUMADIR, node, "numastat")))
            except OSError as e:
                if e.errno == 2:  # No such file or directory
                    continue
                raise
        return numastats

    def get_stats(self):
        # proc.uptime
        procstats_data = {}
        self.f_uptime.seek(0)
        for line in self.f_uptime:
            m = re.match("(\S+)\s+(\S+)", line)
            if m:
                uptime_total = int(float(m.group(1)))
                uptime_idle = int(float(m.group(2)))

        procstats_data['uptime_total'] = uptime_total
        procstats_data['uptime_idle'] = uptime_idle

        # proc.meminfo
        self.f_meminfo.seek(0)
        for line in self.f_meminfo:
            m = re.match("(\w+):\s+(\d+)", line)
            if m:
                procstats_data["meminfo_" + m.group(1).lower()] = int(m.group(2))

        # proc.vmstat
        self.f_vmstat.seek(0)
        for line in self.f_vmstat:
            m = re.match("(\w+)\s+(\d+)", line)
            if not m:
                continue
            if m.group(1) in ("pgpgin", "pgpgout", "pswpin", "pswpout", "pgfault", "pgmajfault"):
                procstats_data["vmstat_" + m.group(1)] = int(m.group(2))

        # proc.stat
        self.f_stat.seek(0)
        for line in self.f_stat:
            m = re.match("(\w+)\s+(.*)", line)
            if not m:
                continue
            if m.group(1) == "cpu":
                fields = m.group(2).split()
                procstats_data['stat_cpu_user'] = int(fields[0])
                procstats_data['stat_cpu_nice'] = int(fields[1])
                procstats_data['stat_cpu_system'] = int(fields[2])
                procstats_data['stat_cpu_idle'] = int(fields[3])
                procstats_data['stat_cpu_iowait'] = int(fields[4])
                procstats_data['stat_cpu_irq'] = int(fields[5])
                procstats_data['stat_cpu_softirq'] = int(fields[6])
                procstats_data['stat_cpu_guest'] = int(fields[7])
                procstats_data['stat_cpu_guest_nice'] = int(fields[8])

            elif m.group(1) == "intr":
                procstats_data['stat_intr'] = int(m.group(2).split()[0])
            elif m.group(1) == "ctxt":
                procstats_data['stat_ctxt'] = int(m.group(2))
            elif m.group(1) == "processes":
                procstats_data['stat_processes'] = int(m.group(2))
            elif m.group(1) == "procs_blocked":
                procstats_data['stat_procs_blocked'] = int(m.group(2))

        self.f_loadavg.seek(0)
        for line in self.f_loadavg:
            m = re.match("(\S+)\s+(\S+)\s+(\S+)\s+(\d+)/(\d+)\s+", line)
            if not m:
                continue

            procstats_data['loadavg_1min'] = float(m.group(1))
            procstats_data['loadavg_5min'] = float(m.group(2))
            procstats_data['loadavg_15min'] = float(m.group(3))
            procstats_data['loadavg_runnable'] = float(m.group(4))
            procstats_data['loadavg_total_threads'] = float(m.group(5))

        return ProcstatsCounters(procstats_buffer=procstats_data)



'''
Colmet-node User Interface
'''
import logging
import argparse
import signal
import sys
import copy
import time

from colmet import VERSION
from colmet.common.backends.zeromq import ZMQInputBackend
from colmet.common.backends.base import StdoutBackend
from colmet.common.exceptions import Error, NoneValueError


LOG = logging.getLogger()


class Task(object):

    def __init__(self, name, options):
        self.name = name
        self.options = options
        self.output_backends = []
        self.init_output_backends()
        self.input_backend = ZMQInputBackend(self.options)
        self.counters_list = []
        self.buffer_size = self.options.buffer_size

    def init_output_backends(self):
        for backend in self.output_backends:
            backend.close()
            del backend
        self.output_backends[:] = []
        options = copy.deepcopy(self.options)
        if options.hdf5_filepath is not None:
            if not options.hdf5_filepath.endswith(".hdf5"):
                options.hdf5_filepath = "%s.%s.hdf5" % \
                    (options.hdf5_filepath, int(time.time()))
            from colmet.collector.hdf5 import HDF5OutputBackend
            self.output_backends.append(HDF5OutputBackend(options))
        if self.options.enable_stdout_backend:
            self.output_backends.append(StdoutBackend(options))
        for backend in self.output_backends:
            backend.open()

    def start(self):
        LOG.info("Starting %s" % self.name)
        self.input_backend.open()
        signal.signal(signal.SIGINT, self.terminate)
        signal.signal(signal.SIGTERM, self.terminate)
        signal.signal(signal.SIGHUP, self.reload)
        try:
            self.loop()
        except:
            self.close_backends()
            raise

    def push(self):
        for backend in self.output_backends:
            if len(self.counters_list) > 0:
                try:
                    backend.push(self.counters_list)
                    LOG.debug("%s metrics has been pushed with %s"
                              % (len(self.counters_list),
                                 backend.get_backend_name()))
                except (NoneValueError, TypeError):
                    LOG.debug("Values for metrics are not there.")
        del self.counters_list[:]

    def close_backends(self):
        self.input_backend.close()
        self.push()
        for backend in self.output_backends:
            backend.close()

    def terminate(self, *args, **kwargs):
        LOG.info("Terminating %s" % self.name)
        self.close_backends()
        sys.exit(0)

    def reload(self, *args, **kwargs):
        LOG.info("Reloading %s" % self.name)
        self.push()
        self.init_output_backends()

    def sleep(self):
        # absolute time is used and based on seconds since 1970-01-01 00:00:00
        # UTC
        now = time.time()
        time_towait = (((now // self.options.sampling_period) + 1) *
                       self.options.sampling_period - now)
        time.sleep(time_towait)

    def loop(self):
        while True:
            now = time.time()
            LOG.debug("Gathering the metrics")
            self.counters_list = \
                self.input_backend.pull(self.options.buffer_size)

            LOG.debug("%s metrics has been pulled from zeromq" %
                      len(self.counters_list))

            LOG.debug("time to take measure: %s sec" % (time.time() - now))

            self.push()
            # sleep to next sampling
            self.sleep()

#
# Main program
#

DESCRIPTION = '''Display and/or collect cpu, memory and i/o bandwidth used
by the processes in a cpuset or a cgroup.'''


def main():
    formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(description=DESCRIPTION,
                                     formatter_class=formatter)

    parser.add_argument('--version', action='version',
                        version='colmet version %s' % VERSION)

    parser.add_argument('-v', '--verbose', action='count', dest="verbosity",
                        default=1)

    parser.add_argument('-s', '--sample-period', type=float,
                        dest='sampling_period', default=5,
                        help='Sampling period of measuring in seconds')

    parser.add_argument('--buffer-size', dest='buffer_size', default=100,
                        help='Defines the maximum number of counters that '
                             'colmet should queue in memory before pushing '
                             'it to output backend', type=int)

    parser.add_argument("--enable-stdout-backend", action='store_true',
                        help='Prints the metrics on STDOUT',
                        dest='enable_stdout_backend', default=False)

    group = parser.add_argument_group('Zeromq')

    group.add_argument("--zeromq-bind-uri", dest='zeromq_bind_uri',
                       help="ZeroMQ bind URI",
                       default='tcp://0.0.0.0:5556')
    group.add_argument("--zeromq-hwm", type=int,
                       default=1000, dest='zeromq_hwm',
                       help="The high water mark is a hard limit on the"
                            " maximum number of outstanding messages "
                            " ZeroMQ shall queue in memory. The value "
                            " of zero means \"no limit\".")

    group.add_argument("--zeromq-linger", type=int,
                       default=0, dest='zeromq_linger',
                       help="Set the linger period for the specified socket."
                            " The value of -1 specifies an infinite linger"
                            " period. The value of 0 specifies no linger"
                            " period.  Positive values specify an upper bound"
                            " for the  linger period in milliseconds.")

    group = parser.add_argument_group('HDF5')

    group.add_argument("--hdf5-filepath", dest='hdf5_filepath', default=None,
                       help='The file path used to store the hdf5 data')

    group.add_argument("--hdf5-complevel", type=int, dest='hdf5_complevel',
                       help='Specifies a compression level for data. '
                            'The allowed range is 0-9. A value of 0 '
                            '(the default) disables compression.',
                       default=0)

    group.add_argument("--hdf5-complib", dest='hdf5_complib', default='zlib',
                       help='Specifies the compression library to be used. '
                            '"zlib" (the default), "lzo", "bzip2" and "blosc" '
                            'are supported.')

    args = parser.parse_args()

    if (args.hdf5_filepath is None and args.enable_stdout_backend is False):
        parser.error("You need to provide at least one output backend "
                     "[hdf5|stdout]")

    # Set the logging value (always display CRITICAL and ERROR)
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s - %(message)s',
        datefmt='%d/%m/%Y %H:%M:%S',
        level=40 - args.verbosity * 10,
    )

    # run
    try:
        Task(sys.argv[0], args).start()
    except KeyboardInterrupt:
        pass
    except Error as err:
        err.show()
        sys.exit(1)
    except Exception as err:
        msg = "Error not handled '%s'" % err.__class__.__name__
        logging.critical(msg)
        if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
            print(repr(err))
        raise


if __name__ == '__main__':
    main()



'''
Netlink message generation/parsing

Copyright 2007        Johannes Berg <johannes@sipsolutions.net>

GPLv2+; See copying for details.
'''

import os
import socket
import struct

try:
    # try to use python 2.5's netlink support
    _dummysock = socket.socket(socket.AF_NETLINK, socket.SOCK_RAW, 0)
    _dummysock.bind((0, 0))
    del _dummysock
    def _nl_bind(descriptor, addr):
        descriptor.bind(addr)
    def _nl_getsockname(descriptor):
        return descriptor.getsockname()
    def _nl_send(descriptor, msg):
        descriptor.send(msg)
    def _nl_recv(descriptor, bufs=16384):
        return descriptor.recvfrom(bufs)
except socket.error:
    # or fall back to the _netlink C module
    try:
        import _netlink
        def _nl_bind(descriptor, addr):
            _netlink.bind(descriptor.fileno(), addr[1])
        def _nl_getsockname(descriptor):
            return _netlink.getsockname(descriptor.fileno())
        def _nl_send(descriptor, msg):
            _netlink.send(descriptor.fileno(), msg)
        def _nl_recv(descriptor, bufs=16384):
            return _netlink.recvfrom(descriptor.fileno(), bufs)
    except ImportError:
        # or fall back to the ctypes module
        import ctypes

        libc = ctypes.CDLL(None)

        class SOCKADDR_NL(ctypes.Structure):
            _fields_ = [("nl_family", ctypes.c_ushort),
                        ("nl_pad",    ctypes.c_ushort),
                        ("nl_pid",    ctypes.c_int),
                        ("nl_groups", ctypes.c_int)]

        def _nl_bind(descriptor, addr):
            addr = SOCKADDR_NL(socket.AF_NETLINK, 0, os.getpid(), 0)
            return libc.bind(descriptor.fileno(),
                             ctypes.pointer(addr),
                             ctypes.sizeof(addr))

        def _nl_getsockname(descriptor):
            addr = SOCKADDR_NL(0, 0, 0, 0)
            len = ctypes.c_int(ctypes.sizeof(addr));
            libc.getsockname(descriptor.fileno(),
                             ctypes.pointer(addr),
                             ctypes.pointer(len))
            return addr.nl_pid, addr.nl_groups;

        def _nl_send(descriptor, msg):
            return libc.send(descriptor.fileno(), msg, len(msg), 0);

        def _nl_recv(descriptor, bufs=16384):
            addr = SOCKADDR_NL(0, 0, 0, 0)
            len = ctypes.c_int(ctypes.sizeof(addr))
            buf = ctypes.create_string_buffer(bufs)

            r = libc.recvfrom(descriptor.fileno(),
                              buf, bufs, 0,
                              ctypes.pointer(addr), ctypes.pointer(len))

            ret = ctypes.string_at(ctypes.pointer(buf), r)
            return ret, (addr.nl_pid, addr.nl_groups)


# flags
NLM_F_REQUEST  = 1
NLM_F_MULTI    = 2
NLM_F_ACK      = 4
NLM_F_ECHO     = 8

# types
NLMSG_NOOP     = 1
NLMSG_ERROR    = 2
NLMSG_DONE     = 3
NLMSG_OVERRUN  = 4
NLMSG_MIN_TYPE = 0x10

class Attr:
    def __init__(self, attr_type, data, *values):
        self.type = attr_type
        if len(values):
            self.data = struct.pack(data, *values)
        else:
            self.data = data

    def _dump(self):
        hdr = struct.pack("HH", len(self.data)+4, self.type)
        length = len(self.data)
        pad = ((length + 4 - 1) & ~3 ) - length
        return hdr + self.data + '\0' * pad

    def __repr__(self):
        return '<Attr type %d, data "%s">' % (self.type, repr(self.data))

    def u16(self):
        return struct.unpack('H', self.data)[0]
    def s16(self):
        return struct.unpack('h', self.data)[0]
    def u32(self):
        return struct.unpack('I', self.data)[0]
    def s32(self):
        return struct.unpack('i', self.data)[0]
    def str(self):
        return self.data
    def nulstr(self):
        return self.data.split('\0')[0]
    def nested(self):
        return parse_attributes(self.data)

class StrAttr(Attr):
    def __init__(self, attr_type, data):
        Attr.__init__(self, attr_type, "%ds" % len(data), data)

class NulStrAttr(Attr):
    def __init__(self, attr_type, data):
        Attr.__init__(self, attr_type, "%dsB" % len(data), data, 0)

class U32Attr(Attr):
    def __init__(self, attr_type, val):
        Attr.__init__(self, attr_type, "I", val)

class U8Attr(Attr):
    def __init__(self, attr_type, val):
        Attr.__init__(self, attr_type, "B", val)

class Nested(Attr):
    def __init__(self, attr_type, attrs):
        self.attrs = attrs
        self.type = attr_type

    def _dump(self):
        contents = []
        for attr in self.attrs:
            contents.append(attr._dump())
        contents = ''.join(contents)
        length = len(contents)
        hdr = struct.pack("HH", length+4, self.type)
        return hdr + contents

NETLINK_ROUTE          = 0
NETLINK_UNUSED         = 1
NETLINK_USERSOCK       = 2
NETLINK_FIREWALL       = 3
NETLINK_INET_DIAG      = 4
NETLINK_NFLOG          = 5
NETLINK_XFRM           = 6
NETLINK_SELINUX        = 7
NETLINK_ISCSI          = 8
NETLINK_AUDIT          = 9
NETLINK_FIB_LOOKUP     = 10
NETLINK_CONNECTOR      = 11
NETLINK_NETFILTER      = 12
NETLINK_IP6_FW         = 13
NETLINK_DNRTMSG        = 14
NETLINK_KOBJECT_UEVENT = 15
NETLINK_GENERIC        = 16

class Message:
    def __init__(self, msg_type, flags=0, seq=-1, payload=None):
        self.type = msg_type
        self.flags = flags
        self.seq = seq
        self.pid = -1
        payload = payload or []
        if isinstance(payload, list):
            contents = []
            for attr in payload:
                contents.append(attr._dump())
            self.payload = ''.join(contents)
        else:
            self.payload = payload

    def send(self, conn):
        if self.seq == -1:
            self.seq = conn.seq()

        self.pid = conn.pid
        length = len(self.payload)

        hdr = struct.pack("IHHII", length + 4*4, self.type,
                          self.flags, self.seq, self.pid)
        conn.send(hdr + self.payload)

    def __repr__(self):
        return '<netlink.Message type=%d, pid=%d, seq=%d, flags=0x%x "%s">' % (
            self.type, self.pid, self.seq, self.flags, repr(self.payload))

class Connection:
    def __init__(self, nltype, groups=0, unexpected_msg_handler=None):
        self.descriptor = socket.socket(socket.AF_NETLINK,
                                        socket.SOCK_RAW, nltype)
        self.descriptor.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 65536)
        self.descriptor.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 65536)
        _nl_bind(self.descriptor, (0, groups))
        self.pid, self.groups = _nl_getsockname(self.descriptor)
        self._seq = 0
        self.unexpected = unexpected_msg_handler
    def send(self, msg):
        _nl_send(self.descriptor, msg)
    def recv(self):
        contents, (nlpid, nlgrps) = _nl_recv(self.descriptor)
        # XXX: python doesn't give us message flags, check
        #      len(contents) vs. msglen for TRUNC
        msglen, msg_type, flags, seq, pid = struct.unpack("IHHII",
                                                          contents[:16])
        msg = Message(msg_type, flags, seq, contents[16:])
        msg.pid = pid
        if msg.type == NLMSG_ERROR:
            errno = -struct.unpack("i", msg.payload[:4])[0]
            if errno != 0:
                err = OSError("Netlink error: %s (%d)" % (
                                                     os.strerror(errno), errno))
                err.errno = errno
                raise err
        return msg
    def seq(self):
        self._seq += 1
        return self._seq

def parse_attributes(data):
    attrs = {}
    while len(data):
        attr_len, attr_type = struct.unpack("HH", data[:4])
        attrs[attr_type] = Attr(attr_type, data[4:attr_len])
        attr_len = ((attr_len + 4 - 1) & ~3 )
        data = data[attr_len:]
    return attrs



class BaseBackend(object):

    def __init__(self, options):
        self.options = options

    def open(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def __del__(self):
        try:
            self.close()
        except:
            pass

    def get_backend_name(self):
        return self.__backend_name__


class OutputBaseBackend(BaseBackend):

    def __init__(self, options):
        super(OutputBaseBackend, self).__init__(options)

    def push(self, msg):
        raise NotImplementedError()


class InputBaseBackend(BaseBackend):

    def __init__(self, options):
        super(InputBaseBackend, self).__init__(options)
        self.job_id_list = []

    def pull(self, request, timestamp):
        raise NotImplementedError()

    def get_counters_class(self, timestamp):
        raise NotImplementedError()


class StdoutBackend(OutputBaseBackend):
    '''
    stdout backend class
    '''
    __backend_name__ = "stdout"

    def open(self):
        pass

    def close(self):
        pass

    def push(self, counters_list):
        '''
        put the metrics to the output backend
        '''
        for counters in counters_list:
            print(counters)



#!/usr/bin/env python
# *-* coding: utf-8 *-*
import re
import os.path as op
from setuptools import setup, find_packages


here = op.abspath(op.dirname(__file__))


requirements = [
    'tables',
    'pyinotify',
    'pyzmq',
]


def read(fname):
    ''' Return the file content. '''
    with open(op.join(here, fname)) as fd:
        return fd.read()


def get_version():
    return re.compile(r".*__version__ = '(.*?)'", re.S)\
             .match(read(op.join(here, 'colmet', '__init__.py'))).group(1)


setup(
    name="colmet",
    version=get_version(),
    description=("A utility to monitor the jobs ressources in a HPC"
                 " environment, espacially OAR"),
    keywords="monitoring, taskstat, oar, hpc, sciences",
    maintainer='Salem Harrache',
    maintainer_email='salem.harrache@inria.fr',
    author="Philippe Le Brouster, Olivier Richard",
    author_email="philippe.le-brouster@imag.fr, olivier.richard@imag.fr",
    url="http://oar.imag.fr/",
    packages=find_packages(),
    long_description=read('README.rst') + '\n\n' + read('CHANGES'),
    install_requires=requirements,
    platforms=['Linux'],
    license="GNU GPL",
    entry_points={
        'console_scripts': [
            'colmet-node = colmet.node.main:main',
            'colmet-collector = colmet.collector.main:main',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Clustering',
        'Programming Language :: Python :: 2.7'
    ]
)



import logging

LOG = logging.getLogger()

from .base import UInt64, BaseCounters


class InfinibandstatsCounters(BaseCounters):
    __metric_name__ = 'infinibandstats_default'

    counters_infinibandstats = {
        # 'key': ( offset,length, type, repr, acc )
        'portXmitData': (UInt64(), 'bytes', 'none', 'portXmitData'),
        'portRcvData': (UInt64(), 'bytes', 'none', 'portRcvData'),
        'portXmitPkts': (UInt64(), 'count', 'none', 'portXmitPkts'),
        'portRcvPkts': (UInt64(), 'count', 'none', 'portRcvPkts')
        }
        
    counters_infinibandstats_to_get = [
        'portXmitData',
        'portRcvData',
        'portXmitPkts',
        'portRcvPkts'
    ]

    _counters = []
        
    for c_name in counters_infinibandstats_to_get:
        (c_type, c_repr, c_acc, c_descr) = counters_infinibandstats[c_name]
        _counters.append((c_name, c_type, c_repr, c_acc, c_descr))

    @classmethod
    def get_zero_counters(cls):
        return cls(infinibandstats_buffer=None, raw=None)

    @classmethod
    def fetch(cls, infinibandstats_backend):
        return infinibandstats_backend.get_infinibandstats()

    def __init__(self, infinibandstats_buffer=None, raw=None):
        BaseCounters.__init__(self, raw=raw)
        if raw is not None:
            pass
        elif infinibandstats_buffer is None:
            self._empty_fill()
        else:
            for name in InfinibandstatsCounters._counter_definitions:
                self._counter_values[name] = infinibandstats_buffer[name]



import os
import re

import glob

from colmet.common.backends.base import InputBaseBackend
from colmet.common.job import Job
from colmet.common.metrics.lustrestats import LustrestatsCounters


class LustrestatsBackend(InputBaseBackend):
    __backend_name__ = "lustrestats"

    def open(self):
        self.lustrestats = LustreStats(self.options)
        # job with id equal to 0 is the fictive job to gather nodes' monitoring
        # measures
        self.job_0 = Job(self, 0, self.options)

    def close(self):
        pass

    def get_lustrestats(self):
        counters = self.lustrestats.get_stats()
        return counters

    def get_counters_class(self):
        return LustrestatsCounters

    def pull(self):
        self.job_0.update_stats()
        return self.job_0.get_stats()


class LustreStats(object):
    '''
    Read counters from all mounted lustre fs
    from the file stats under the directories:
  
    /proc/fs/lustre/llite/lustre-xxxx/stats

    From the file stat we use 2 entries:

    read_bytes          17996 samples [bytes] 0 4194304 30994606834
    write_bytes         9007 samples [bytes] 2 4194304 31008331389

    First number = number of times (samples) the OST has handled a read or write.
    Second number = the minimum read/write size
    Third number = maximum read/write size
    Fourth = sum of all the read/write requests in bytes, the quantity of data read/written.
    '''

    def __init__(self, option):
        self.options = option

    def get_stats(self):

        lustre_nb_read = 0
        lustre_bytes_read = 0
        lustre_nb_write = 0
        lustre_bytes_write = 0
        
        for stats_file_name in glob.glob('/proc/fs/lustre/llite/lustre-*/stats'):
            with open(stats_file_name) as stats_file:
                for line in stats_file:
                    if re.match('read_bytes', line):
                        values = line.split()
                        lustre_nb_read += int(values[1])
                        lustre_bytes_read += int(values[-1])
                    if re.match('write_bytes', line):
                        values = line.split()
                        lustre_nb_write += int(values[1])
                        lustre_bytes_write += int(values[-1])
                        
        lustrestats_data = {'lustre_nb_read': lustre_nb_read, 'lustre_bytes_read': lustre_bytes_read,
                            'lustre_nb_write': lustre_nb_write, 'lustre_bytes_write': lustre_bytes_write
        }
                        
        return LustrestatsCounters(lustrestats_buffer=lustrestats_data)



import asyncore
import logging
import os
from threading import Thread
import time

from pyinotify import ProcessEvent, AsyncNotifier, WatchManager
import pyinotify

LOG = logging.getLogger()


class ThreadProperty(object):
    """Decorator that make any method a thread object."""
    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        _thread = obj.__dict__.get(self.__name__)
        if _thread is None:
            _thread = Thread(target=self.func, args=(obj,))
            # thread.daemon causes the thread to terminate when the main
            # process ends
            _thread.daemon = True
            obj.__dict__[self.__name__] = _thread
        return _thread


as_thread = ThreadProperty


def wait_dir(path):
    """Wait until the `path` folder creation."""
    while True:
        if os.path.exists(path) and os.path.isdir(path):
            break
        time.sleep(5)


class AsyncFileNotifier(object):
    def __init__(self, paths, callback):
        class EventHandler(ProcessEvent):
            def process_IN_CREATE(self, event):  # noqa
                callback()

            def process_IN_DELETE(self, event):  # noqa
                callback()

        manager = WatchManager()  # Watch Manager
        mask = pyinotify.IN_DELETE | pyinotify.IN_CREATE  # watched events
        self.async_notifier = AsyncNotifier(manager, EventHandler())
        # aggregate inotify events
        try:
            self.async_notifier.coalesce_events()
        except AttributeError as inst:
            LOG.warn('Can not coalesce events, pyinotify does not seem to '
                     'support it (maybe too old): %s' % inst)
        for path in paths:
            if os.path.exists(path):
                manager.add_watch(path, mask, rec=False)
            else:
                LOG.warn("%s folder doesn't exist yet." % path)
                wait_dir(path)
                manager.add_watch(path, mask, rec=False)
                LOG.info("%s has just been created." % path)

    def loop(self):
        LOG.info("Starting inotify daemon.")
        asyncore.loop()



import os
import re

from subprocess import check_output

from colmet.common.backends.base import InputBaseBackend
from colmet.common.job import Job
from colmet.common.metrics.infinibandstats import InfinibandstatsCounters



class InfinibandstatsBackend(InputBaseBackend):
    __backend_name__ = "infinibandstats"

    def open(self):
        self.infinibandstats = InfinibandStats(self.options)
        # job with id equal to 0 is the fictive job to gather nodes' monitoring
        # measures
        self.job_0 = Job(self, 0, self.options)

    def close(self):
        pass

    def get_infinibandstats(self):
        counters = self.infinibandstats.get_stats()
        return counters

    def get_counters_class(self):
        return InfinibandstatsCounters

    def pull(self):
        self.job_0.update_stats()
        return self.job_0.get_stats()


class InfinibandStats(object):

    def __init__(self, option):
        self.options = option

    def get_stats(self):
        
        infinibandstats_data = {}
        
        # perfquery = check_output(["/usr/sbin/perfquery", "-x"])
        perfquery = check_output(["/usr/sbin/perfquery"])
        
        m = re.search(r'PortXmitData:\.*(\d+)', perfquery)
        if m:
            infinibandstats_data['portXmitData'] = 4 * int(m.group(1))
        m = re.search(r'PortRcvData:\.*(\d+)', perfquery)
        if m:
            infinibandstats_data['portRcvData'] = 4 * int(m.group(1))
        m = re.search(r'PortXmitPkts:\.*(\d+)', perfquery)
        if m:
            infinibandstats_data['portXmitPkts'] = int(m.group(1))
        m = re.search(r'PortRcvPkts:\.*(\d+)', perfquery)
        if m:
            infinibandstats_data['portRcvPkts'] = int(m.group(1))
                        
        return InfinibandstatsCounters(infinibandstats_buffer=infinibandstats_data)
